/* Copyright 2017, 2018 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Search all trees using gentreeg for even n with > 2^(n/2) minimal domsets.
   Slower than minimal-domsets-max-even2.c which prunes the search early
   when 2^(n/2) is unreachable.

   n=22  5 sec

   n=24  36 sec
   max minimal_domsets=4096 count 551
   tree count 39299897
   vpar_count_free(12) == 551
   vpar_count_free(24) == 39299897

   n=26  270 sec = 4:30 mins
   6 end, tree count 279793450
   max minimal_domsets=8192 count 1301
   vpar_count_free(13) == 1301
   vpar_count_free(26) == 279793450

   n=28  44:14 mins
   tree count 2023443032
   max minimal_domsets=16384 count 3159
   vpar_count_free(14) == 3159
   vpar_count_free(28) == 2023443032

   n=30  5:09 hrs
   max minimal_domsets=32768 count 7741
   vpar_count_free(30) == 14 830 871 802
   44.5*vpar_count_free(30)/vpar_count_free(28) / 60
   est 5.5 hours

   n=32
   vpar_count_free(30) == 109 972 410 221
   5.2*vpar_count_free(32)/vpar_count_free(30)
   found n=32 0,1,2,3,4,5,4,7,4,9,4,11,2,13,2,15,2,17,1,19,20,21,20,23,20,25,1,27,1,29,1,31
   count 65960 > target 65536
   :__`abcbebgbi`k`m`o_qrsrurw_y_{_}
   n=32 new high max_value 65960 target 65536  at tree_count 104760057687
   n=32 end, tree count 109972410221
   max minimal_domsets=65960 count 1
   elapsed time 143748 = 2395:48 = 40 hours
   https://hog.grinvin.org/ViewGraphInfo.action?id=30700

   n=34
   44.5*vpar_count_free(34)/vpar_count_free(28) / 60 / 24
   (est 12 days)
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int my_geng_main(int argc, char**argv);
int my_gentreeg_main(int argc, char**argv);

const int option_verbose = 1;

unsigned long long  trees_total_domnum[64];

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))


unsigned long  target_count;
unsigned long  good_num = 0;
unsigned long long  tree_count = 0;
unsigned long  max_value = 0;
int            max_count = 0;

void
my_gentreeg_outproc (FILE *outfile, int *vpar, int n)
{
  DEBUG(printf("my_gentreeg_outproc() n=%d\n", n));
  tree_count++;

  unsigned long  count_minimal_domsets
    = nautyextra_minimal_domsets_count_vpar(vpar,n);

  if (count_minimal_domsets > target_count) {
    printf("found "); nautyextra_print_vpar(vpar,n);
    printf("count %lu > target %lu\n", count_minimal_domsets, target_count);
    printf("%s", nautyextra_vpar_to_s6(vpar,n));
    printf("--\n");
  }

  if (count_minimal_domsets > max_value) {
    max_value = count_minimal_domsets;
    max_count = 0;
#if 1
    printf("n=%d new high max_value %lu target %lu  at tree_count %Lu\n",
           n, max_value, target_count, tree_count);
#endif
  }
  if (max_value == count_minimal_domsets) {
    max_count++;
  }
}

int
main (void)
{
  setbuf(stdout,NULL);
  nautyextra_want_progress = 1;

  int num_vertices;
  for (num_vertices = 32;
       num_vertices <= 32;
       num_vertices += 2) {
    tree_count = 0;
    target_count = 1 << (num_vertices/2);
    max_value = 0;
    max_count = 0;
    time_t start_time;
    time(&start_time);

    printf("gentreeg  num_vertices=%d target_count=%lu\n",
           num_vertices, target_count);
    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);

    int gentreeg_argc = 0;
    char *gentreeg_argv[10];
    gentreeg_argv[gentreeg_argc++] = "gentreeg"; /* progname */
    gentreeg_argv[gentreeg_argc++] = "-q";       /* quiet */
    gentreeg_argv[gentreeg_argc++] = num_vertices_str;
    gentreeg_argv[gentreeg_argc] = NULL;
    my_gentreeg_main(gentreeg_argc,gentreeg_argv);

    time_t end_time;
    time(&end_time);

    printf("n=%d end, tree count %Lu\n", num_vertices, tree_count);
    printf("  max minimal_domsets=%lu count %d\n", max_value, max_count);

    time_t elapsed_time = end_time - start_time;
    printf("  elapsed time %ld = %ld:%02ld\n",
           elapsed_time, elapsed_time/60, elapsed_time%60);
    printf("\n");
  }

  return 0;
}
