/* Copyright 2018, 2022 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/*
  reps                   3
  n=    1 2 3 4 5  6  7  8  9
  conn=[
  disc=[
  \\                       ^^

  most minimum dominating sets in a connected graph
  conn

  most minimum dominating sets in a disconnected graph
  disc

  vector_product(v) = prod(i=1,#v,v[i]);
  disc=vector(#conn+1);  \\ disconnected, by partitions of connected
  for(n=1,#disc, my(m=0); \
  forpart(p=n, if(p!=Vecsmall([n]),\
  m=max(m,vector_product(vecextract(conn,p))))); \
  disc[n]=m); disc

  my(l=List([])); \
  forpart(p=12, listput(l,[vector_product(vecextract(conn,p)),Vec(p)])); vecsort(Vec(l))

  n=2 2 sets, unique
  path-2
  :An

  n=3 3 sets, unique
  3-cycle
  :BcN

  n=4 6 sets, unique
  4-cycle
  :Cci

  n=5 max 9 reps 1
  https://hog.grinvin.org/ViewGraphInfo.action?id=438
  :DaHIQN
  9^(1/5)
  1.551845

  n=6 n max 15 reps 1
  octahedral graph
  https://hog.grinvin.org/ViewGraphInfo.action?id=226
  :Ea@_WGxGs
  15^(1/6)
  1.57041780247501973531052
  A011350

  n=7 connected or disconnected, 22 sets
  https://hog.grinvin.org/ViewGraphInfo.action?id=868
  :FaqiBQRP^
  22^(1/7)
  1.555158

  --
  n=9 connected max 51 reps 1
  261080 graphs, or 197772 graphs degree>=2
  :H`A_wa_K`WaEQG`cJ
  cf minimals has 2 others equal
  https://hog.grinvin.org/ViewGraphInfo.action?id=30667

  --
  n=10 connected max 87 reps 1
  count 11716571 graphs, 9808209 degree>=2
  87^(1/10) = 1.562974618
  9808209.0/197772 = 49
  :IcAKWQJGhhcLK`s^
  https://hog.grinvin.org/ViewGraphInfo.action?id=30675

  n=10 any 90 sets 4-cycle and octahedral n=6
  count 12005168 graphs
  :I`AGicgU`YENSps^
  https://hog.grinvin.org/ViewGraphInfo.action?id=30677

*/


#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int my_geng_main(int argc, char**argv);
int my_gentreeg_main(int argc, char**argv);

int option_verbose = 0;


/* --------------------------------------------------------------------------- */

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))
#define numberof(array)  (sizeof(array)/sizeof((array)[0]))


unsigned long long graphs_count;
double b_max;
int b_max_n;
int b_max_count;

int  max_count;
int  max_reps;
char *max_s6[10];

void
my_geng_outproc (FILE *outfile, graph *g, int n)
{
  const int m = 1;
  graphs_count++;

  if (NAUTYEXTRA_PROGRESS)
    printf("n=%d graphs %Lu  (reps %d)\r",
           n, graphs_count, nautyextra_progress_reps);

  struct nautyextra_domnum_and_count  info;
  nautyextra_domnum_and_count1(g,n, &info);

  double b = pow(info.count, 1.0/n);
  if (b>b_max) {
    b_max = b;
    printf ("b new high at graph %Lu n=%d pow %.9f\n", graphs_count, n, b);
    printf("%s%s", SPARSE6_HEADER, nautyextra_canonical_str(g,m,n,1));
    b_max_n = n;
    b_max_count = info.count;
  }

  if (info.count > max_count) {
    max_count = info.count;
    max_reps = 0;
  }
  if (max_count == info.count) {
    if (max_reps < numberof(max_s6)) {
      if (max_s6[max_reps]) free(max_s6[max_reps]);
      max_s6[max_reps] = strdup(nautyextra_canonical_str(g,m,n,1));
    }
    max_reps++;
  }
}

int
main (void)
{
  nautyextra_want_progress = 1;
  setbuf(stdout,NULL);

  /* if (0) { */
  /*   try_octahedral_connect (); */
  /*   return 0; */
  /* } */

  int num_vertices;
  for (num_vertices = 1;
       num_vertices <= 11;
       num_vertices++) {
    if (option_verbose >= 1) {
      printf ("n=%d geng\n", num_vertices);
    }
    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);

    graphs_count = 0;
    max_count = 0;
    max_reps = 0;
    /* b_max = 0.0; */
    int geng_argc = 0;
    char *geng_argv[10];
    geng_argv[geng_argc++] = "geng";
    geng_argv[geng_argc++] = "-q";     /* quiet */
    geng_argv[geng_argc++] = "-c";     /* connected graphs */
    /* geng_argv[geng_argc++] = "-d2";    /\* minimum degree *\/ */
    geng_argv[geng_argc++] = num_vertices_str;
    geng_argv[geng_argc] = NULL;
    my_geng_main(geng_argc,geng_argv);

    printf ("n=%d count %Lu graphs\n", num_vertices, graphs_count);
    printf ("  maximum %.9f = %d^(1/%d)\n",
            b_max, b_max_count, b_max_n);
    printf ("  this n max %d reps %d\n",
            max_count, max_reps);
    int i;
    for(i = 0; i < max_reps; i++)
      printf("%s", max_s6[i]);

    for(i = 0; i < max_reps; i++) {
      graph g[32];
      const char *str = max_s6[i];
      int n = graphsize((char*)str);
      assert(n == num_vertices);
      int m = SETWORDSNEEDED(n);
      assert(m==1);
      stringtograph((char*)str, g,m);
      printf(" cf minimals %lu\n", nautyextra_minimal_domsets_count1(g,n));
    }
    printf ("\n");
  }
}
