/* Copyright 2017, 2018, 2019, 2022 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

#define NAUTY_CPU_DEFINED 1  /* not CPUDEFS */
#include <assert.h>
#include <stdint.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/naugroup.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

#define NAUTYEXTRA_SET1_LOOP(ss,n,body)                 \
  do {                                                  \
    assert (n >= 1);                                    \
    if (! (SETWORDSNEEDED(n) == 1)) abort();            \
    setword __NAUTYEXTRA_SET_LOOP__inc = BITT[n-1];     \
    (ss) = 0;                                           \
    do {                                                \
      body;                                             \
    } while ((ss) += __NAUTYEXTRA_SET_LOOP__inc);       \
  } while (0)

#define numberof(array)  (sizeof(array)/sizeof((array)[0]))

int
compare_int (const void *p, const void *q)
{
  int pp = *(int*)p;
  int qq = *(int*)q;
  int d = pp - qq;
  return (d<0 ? -1 : d==0 ? 0 : 1);
}

void
nautyextra_print_array (const int *array, int n)
{
  int i;
  for (i = 0; i < n; i++) {
    if (i) printf(",");
    printf("%d",array[i]);
  }
  printf("\n");
}

/* ------------------------------------------------------------------------- */

void
try_generalized_Petersen_count (void)
{
  /* A077105 = number of non-isomorphic k for given n */
  static const int A077105[] = {
    0,0,0, /* 0 to 2 */
    1, 1, 2, 2, 2, 3, 3, 4, 3, 5, 4, 5, 6, 6, 5, 7, 5, 8, 8, 8, 6, 11, 8,
    10, 9, 11, 8, 13, 8, 12, 12, 13, 12, 15, 10, 14, 14, 17, 11, 18, 11,
    17, 17, 17, 12, 21, 14, 20, 18, 20, 14, 22, 18, 23, 20, 22, 15, 27, 16,
    23, 23, 24, 22, 28, 17, 26, 24, 29, 18, 31, 19, 28, 28, 29, 24, 33, 20
  };
  struct seen {
    char *str;
    int  k;
  };
  int N_max = 120;
  int N;
  for (N=3; N<=N_max; N++) {
    DYNALLSTAT(struct seen,seen,seen_sz);
    DYNALLSTAT(graph,g, g_sz);
    DYNALLSTAT(graph,g2,g2_sz);
    int seen_upto = 0;
    int n = N<<1;
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g, g_sz, m,n,"malloc");
    DYNALLOC2(graph,g2,g2_sz,m,n,"malloc");
    int k;
    /* A077105 goes 1<=k<=floor((n-1)/2) so for N even exclude k=N/2 which is
       crossed lines
    */
    for (k = 1; k <= ((N-1)>>1); k++) {
      nautyextra_make_Petersen (g,m,n,k);
      nautyextra_validate(g,m,n);
      nautyextra_canonise(g,m,n, g2);
      nautyextra_validate(g2,m,n);
      char *str = ntog6(g2,m,n);
      DEBUG(printf("N=%d n=%d,m=%d k=%d %s", N,n,m,k,str));
      int i;
      for (i = 0; i < seen_upto; i++) {
        if (strcmp(str,seen[i].str)==0) {
          DEBUG(printf("  same as i=%d\n", i));
          goto NEXT_K;
        }
      }
      if (! (str = strdup(str))) abort();
      DEBUG(printf("  add seen %d %p\n", seen_upto, str));
      DYNREALLOC(struct seen,seen,seen_sz,seen_upto+1,"realloc");
      seen[seen_upto].k = k;
      seen[seen_upto++].str = str;
    NEXT_K:
      ;
    }
    printf("%d,",seen_upto);
    if (N<numberof(A077105)) {
      if (seen_upto != A077105[N]) {
        printf("\nN=%d got %d but A077105 %d diff %d\n",
               N, seen_upto, A077105[N],
               seen_upto - A077105[N]);
        if (N<10) {
          int i;
          for (i = 0; i < seen_upto; i++) {
            printf("  k=%d %s", seen[i].k, seen[i].str);
          }
        }
      }
    }
    int i;
    for (i = 0; i < seen_upto; i++)
      free(seen[i].str);
  }
  printf("\n");
}

void
try_generalized_Petersen_sets (void)
{
  SG_DECL(sg);
  SG_INIT(sg);
  int bad = 0;

  static const int A182054[] = {
    8,3,39,171,1055,5828,33327,188499,1069855,6065487,34399844,195074223
  };
  static const int A182077[] = {  /* starting N=3 */
    13,76,435,2461,13971,79197,449188,2547179,14445169,81917079,464547653
  };
  int N;
  for (N=3; N<16; N++) {
    DYNALLSTAT(graph,g, g_sz);
    int n = N<<1;
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g, g_sz, m,n,"malloc");
    nautyextra_make_Petersen (g,m,n,2);
    nautyextra_validate(g,m,n);
    nauty_to_sg(g,&sg,m,n);

    {
      unsigned long long  got  = nautyextra_indsets_count_sg(&sg);
      unsigned long long  want = (N&1
                                  ? A182077[(N>>1)-1]    /* odd */
                                  : A182054[(N>>1)]);  /* even */
      if (got != want) {
        bad++;
      }
      printf("N=%d got %Lu want %Lu diff %Ld\n",
             N, got, want, (long long) got - want);
    }
  }
  if (bad) abort();
  SG_FREE(sg);
}

void
try_generalized_Petersen_diameter (void)
{
  int bad = 0;

  static const int A274047[] = {  /* starting N=5 */
    2,4,3,4,4,5,5,5,5,6,6,6,6,7,7,7,7,8,8,8,8,9,9,9,9,10,10,10,10,11,11,11,11,
    12,12,12,12,13,13,13,13,14,14,14,14,15,15,15,15,16,16,16,16,17,17,17,17,18,
    18,18,18,19,19,19,19,20,20,20,20,21,21,21,21,22,22,22,22,23,23,23,23,24,24,
    24,24,25
  };
  int i;
  for (i=0; i<numberof(A274047); i++) {
    int N = i+5;
    DYNALLSTAT(graph,g, g_sz);
    int n = N<<1;
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g, g_sz, m,n,"malloc");
    nautyextra_make_Petersen (g,m,n,2);
    nautyextra_validate(g,m,n);

    int radius,got;
    diamstats(g,m,n, &radius,&got);
    int  want = A274047[i];
    if (got != want) {
      printf("N=%d got %d want %d diff %d\n", N, got, want, got - want);
      bad++;
    }
  }
  if (bad) abort();
}

/* ------------------------------------------------------------------------- */

void
try_cycle_totdomnum (void)
{
SG_DECL(sg);
  SG_INIT(sg);
  const int n = 5;
  const int m = SETWORDSNEEDED(n);
  setword g[n*m];

  nautyextra_make_cycle_sg (&sg,n);
  int m_out;
  sg_to_nauty(&sg,g,m,&m_out);
  printf ("girth    %d\n", girth(g,m,n));
  printf ("edges    %d\n", nautyextra_num_edges(g,m,n));

  struct nautyextra_totdomnum_and_count  info;
  nautyextra_totdomnum_and_count_sg(&sg, &info);
  printf ("totdomnum  %d count %Lu\n", info.totdomnum, info.count);

  nautyextra_view_synchronous = 1;
  nautyextra_view_sg (&sg);
}

/* ------------------------------------------------------------------------- */

void
try_Petersen_7_2 (void)
{
  SG_DECL(sg);
  SG_INIT(sg);
  const int n = 14;
  const int m = SETWORDSNEEDED(n);
  const int k = 2;
  setword g[n*m];
  EMPTYGRAPH(g,m,n);
  nautyextra_make_Petersen (g,m,n,k);

  printf ("girth    %d\n", girth(g,m,n));
  printf ("edges    %d\n", nautyextra_num_edges(g,m,n));

  nauty_to_sg(g,&sg, m,n);
  struct nautyextra_indomnum_and_count  info;
  nautyextra_indomnum_and_count_sg(&sg, &info);
  printf ("indomnum  %d count %Lu\n", info.indomnum, info.count);

  /* nautyextra_view_synchronous = 1; */
  /* nautyextra_view (g,m,n); */

  setword g2[n*m];
  nautyextra_canonise(g,m,n, g2);
  printf("%s", ntog6((graph*)g,m,n));
  printf("%s", ntog6((graph*)g2,m,n));

  printf ("\n");
  const char *str = "Ms?G?DCQ@DAID_IO?";
  printf("%s\n", str);
  SG_DECL(sg2);
  SG_INIT(sg2);
  int nloops;
  stringtosparsegraph((char*)str, &sg2, &nloops);
  printf("%s", sgtog6(&sg2));
  nautyextra_indomnum_and_count_sg(&sg2, &info);
  printf ("indomnum  %d count %Lu\n", info.indomnum, info.count);

  SG_DECL(sg2_can);
  SG_INIT(sg2_can);
  nautyextra_canonise_sg(&sg,&sg2_can);
  printf("%s", sgtog6(&sg2_can));

  printf("isomorphic %d\n", nautyextra_is_isomorphic_sg(&sg,&sg2));
}

/* ------------------------------------------------------------------------- */

/* triangle with attached vertices
   unique maximum indomnum 5 among n=9 graphs

   indnum    6 count 1
   indomnum  5 count 3
   domnum    3 count 1
   totdomnum 3 count 1

   indsets counts 1, 9, 27, 38, 27, 9, 1
   size =         0  1   2   3   4  5  6
   size=5 count 6 of 5-leaf, 3 of 1 triangle 4 leaf
*/
void
try_indomnum_n9 (void)
{
  SG_DECL(sg);
  SG_INIT(sg);
  nautyextra_make_cycle_sg(&sg,3);
  nautyextra_add_edge_sg(&sg,0,3);
  nautyextra_add_edge_sg(&sg,0,4);
  nautyextra_add_edge_sg(&sg,1,5);
  nautyextra_add_edge_sg(&sg,1,6);
  nautyextra_add_edge_sg(&sg,2,7);
  nautyextra_add_edge_sg(&sg,2,8);
  nautyextra_view_sg(&sg);

  const int n = 9;
  int m = SETWORDSNEEDED(n);
  assert(m==1);
  graph g[m*n];

  {
    unsigned long long count = nautyextra_indsets_count_sg(&sg);
    printf ("indsets count %Lu\n", count);
  }
  {
    sg_to_nauty(&sg, g,m,&m);
    assert(m==1);
    struct nautyextra_indnum_and_count  info;
    nautyextra_indnum_and_count(g,m,n, &info);
    printf ("indnum    %d count %Lu\n", info.indnum, info.count);
  }
  {
    struct nautyextra_domnum_and_count  info;
    nautyextra_domnum_and_count_sg(&sg, &info);
    printf ("domnum    %d count %Lu\n", info.domnum, info.count);
  }
  {
    struct nautyextra_indomnum_and_count  info;
    nautyextra_indomnum_and_count_sg(&sg, &info);
    printf ("indomnum  %d count %Lu\n", info.indomnum, info.count);
  }
  {
    struct nautyextra_totdomnum_and_count  info;
    nautyextra_totdomnum_and_count_sg(&sg, &info);
    printf ("totdomnum %d count %Lu\n", info.totdomnum, info.count);
  }

  {
    int count[n+1];
    memset(count,'\0',sizeof(count));
    setword ss;
    NAUTYEXTRA_SET1_LOOP
      (ss,n, count[POPCOUNT(ss)] += nautyextra_is_indset_sg(&sg,&ss));
    int i;
    for(i=0;i<=n;i++)
      printf ("  indsets size %d count %d\n", i, count[i]);
  }
  SG_FREE(sg);
}

void
writeautom (int *p, int n)
{
  int i;
  printf("autom ");
  for (i = 0; i < n; i++)
    printf(" %2d",p[i]);
  printf("\n");
}

void
try_pseudosimilar (void)
{
  static DEFAULTOPTIONS_SPARSEGRAPH(options);
  statsblk stats;

  SG_DECL(sg);
  SG_DECL(canon_sg);

  options.getcanon = TRUE;
  options.userautomproc = groupautomproc;
  options.userlevelproc = grouplevelproc;

  printf("make graph\n");

  if (0) {
    /* two triangles,
       only automorphism is swapping 0<->1 in hanging triangle

       .    2       3
       .   / \     / \
       .  1---5---6---7---4---0

       https://hog.grinvin.org/ViewGraphInfo.action?id=30306
       https://hog.grinvin.org/ViewGraphInfo.action?id=30308  removal
    */
    nautyextra_add_edge_sg(&sg, 0,1);
    nautyextra_add_edge_sg(&sg, 0,2);
    nautyextra_add_edge_sg(&sg, 1,2);
    nautyextra_add_edge_sg(&sg, 2,3);
    nautyextra_add_edge_sg(&sg, 3,4);
    nautyextra_add_edge_sg(&sg, 3,5);
    nautyextra_add_edge_sg(&sg, 4,5);
    nautyextra_add_edge_sg(&sg, 5,6);
    nautyextra_add_edge_sg(&sg, 6,7);
  }
  if (0) {
    /* 3 copies of K4-e

       .  0   3---4   7
       .  |   | / | / |
       .  1---2---5---6
       https://hog.grinvin.org/ViewGraphInfo.action?id=30310
    */
    nautyextra_add_edge_sg(&sg, 0,1);
    nautyextra_add_edge_sg(&sg, 1,2);
    nautyextra_add_edge_sg(&sg, 2,3);
    nautyextra_add_edge_sg(&sg, 2,4);
    nautyextra_add_edge_sg(&sg, 2,5);
    nautyextra_add_edge_sg(&sg, 3,4);
    nautyextra_add_edge_sg(&sg, 5,4);
    nautyextra_add_edge_sg(&sg, 5,7);
    nautyextra_add_edge_sg(&sg, 5,6);
    nautyextra_add_edge_sg(&sg, 6,7);
  }
  if (1) {
    /* n=8 line graph pseudosimilars
       GCQREO   edges= 9
       GCQbUG         10
       GCpddW         12     canonical 5,6 given 2,4
       GEhtr{         17     vertices 4,5
    */
    int nloops;

    stringtosparsegraph("GEhtr{",&sg,&nloops);
    assert (nloops == 0);
    int n = sg.nv;
    assert (n == 8);
    int m = SETWORDSNEEDED(n);
    assert (m == 1);
    graph g[8];
    int m_out;
    sg_to_nauty(&sg,g,m,&m_out);
    assert (m == m_out);
    assert (nautyextra_is_linegraph(g,m,n));
  }

  printf("\n");
  printf("as generated\n");
  nautyextra_print_sg(&sg);
  nautyextra_view_sg (&sg);

  /* DYNALLSTAT(int, degrees, degrees_sz); */
  /* DYNALLOC1(int, degrees, degrees_sz, sg.nv, "malloc"); */
  /* nautyextra_degrees_array_sg(&sg); */
  /* printf("degrees "); nautyextra_print_array (degrees, sg.nv); */

  int num_vertices = sg.nv;
  DYNALLSTAT(int, lab,    lab_sz);
  DYNALLSTAT(int, ptn,    ptn_sz);
  DYNALLSTAT(int, orbits, orbits_sz);
  DYNALLSTAT(int, map,    map_sz);
  DYNALLOC1(int, lab,    lab_sz,    num_vertices, "malloc");
  DYNALLOC1(int, ptn,    ptn_sz,    num_vertices, "malloc");
  DYNALLOC1(int, orbits, orbits_sz, num_vertices, "malloc");
  DYNALLOC1(int, map,    map_sz,    num_vertices, "malloc");

  printf("\nlabel sg\n");
  sparsenauty(&sg,lab,ptn,orbits,&options,&stats,&canon_sg);

  printf("canonical\n");
  printf("lab "); nautyextra_print_array (lab, sg.nv);
  /* nautyextra_degrees_array_sg(&canon_sg); */
  /* printf("degrees "); nautyextra_print_array (degrees, sg.nv); */
  /* nautyextra_print_sg(&canon_sg); */
  /* nautyextra_view_sg (&canon_sg); */

  sortlists_sg(&sg);

  printf("putgraph_sg()\n");
  putgraph_sg(stdout, &sg, 79);  /* line length */

  printf("\n");
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&sg);
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&canon_sg);

  printf("\n");
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&sg);
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&canon_sg);

  printf("aresame  %d\n", aresame_sg(&sg,&canon_sg));

  printf ("\nGenerators\n");
  printf ("automorphism group size = ");
  writegroupsize(stdout, stats.grpsize1, stats.grpsize2);
  printf ("\n");
  printf ("num orbits    %d\n", stats.numorbits);
  printf ("  orbits      ");
  int i;
  for (i = 0; i < sg.nv; i++) printf(" %d",orbits[i]);
  printf("\n");
  printf ("  v =         ");
  for (i = 0; i < sg.nv; i++) printf(" %d",i);
  printf("\n");
  printf ("errstatus     %d\n", stats.errstatus);
  printf ("numnodes      %lu\n", stats.numnodes);
  printf ("numbadleaves  %lu\n", stats.numbadleaves);
  printf ("maxlevel      %d\n", stats.maxlevel);

  {
    grouprec *group = groupptr(FALSE);
    if (! group) {
      printf("no grouprec\n");
    } else {
      printf("grouprec %p\n", group);
      makecosetreps(group);  /* ready for allgroup() */
      allgroup(group,writeautom);
    }
  }

  const int n = 8;
  assert (sg.nv == n);
  int m = SETWORDSNEEDED(n);
  graph g[m*n];
  int m_out;
  sg_to_nauty(&sg,g,m,&m_out);

  struct nautyextra_similar_info info;
  nautyextra_similar(g,m,n, &info);

  printf("similar:\n");
  nautyextra_print_matrix(info.similar,m,n);
  printf("numorbits %d\n", info.numorbits);
  printf("orbit_max %d\n", info.orbit_max);
  printf("\n");

  printf("removal similar:\n");
  nautyextra_print_matrix(info.removal_similar,m,n);
  printf("\n");

  printf("pseudosimilar:\n");
  nautyextra_print_matrix(info.pseudosimilar,m,n);
  printf("pseudosimilar_clique_size %d\n", info.pseudosimilar_clique_size);

  {
    int i,j;
    for (i = 0; i < sg.nv; i++)
      for (j = i+1; j < sg.nv; j++)
        if (ISELEMENT(GRAPHROW(info.pseudosimilar,m,i),j))
          printf("pseudosimilar %d -- %d\n", i,j);
  }
  printf("vertex  ");
  for (i = 0; i < sg.nv; i++) printf(" %d",i);
  printf("\n");
  printf("degree  ");
  for (i = 0; i < sg.nv; i++) printf(" %d",sg.d[i]);
  printf("\n");
  printf("adj2s   ");
  for (i = 0; i < sg.nv; i++) {
    int count_neighbour_2s = 0;
    int pos;
    for (pos = 0; pos < sg.d[i]; pos++) {
      int j = sg.e[sg.v[i]+pos];
      count_neighbour_2s += (sg.d[j]==2);
    }
    printf(" %d",count_neighbour_2s);
  }
  printf("\n");

  printf("adj3s   ");
  for (i = 0; i < sg.nv; i++) {
    int count_neighbour_3s = 0;
    int pos;
    for (pos = 0; pos < sg.d[i]; pos++) {
      int j = sg.e[sg.v[i]+pos];
      count_neighbour_3s += (sg.d[j]==3);
    }
    printf(" %d",count_neighbour_3s);
  }
  printf("\n");
}


/* ------------------------------------------------------------------------- */

void
try_isomorphism_classes (void)
{
  /* Put all trees of N vertices into isomorphism equivalence classes, by
     comparing each to a representative.

     undirected tree isomorphic trees
     1,1,2,3,6,11
     A000055

     directed tree isomorphic trees
     count 1,2,4,9,20,48
     n =   2 3 4
     A000081 unlabelled rooted n nodes
  */

  const int forest = 0;

  const int max_n = 10;
  int vpar[max_n+1];
  int max_reps = 1000;
  int rep_vpar[2][max_reps][max_n+1];

  int n;
  for (n=2; n<=max_n; n++) {
    printf("n=%d\n", n);
    int count_arrays = 0;
    int count_trees = 0;
    int num_reps[2];
    num_reps[0] = 0;
    num_reps[1] = 0;

    if (forest) {
      memset(vpar,'\0',(n+1)*sizeof(vpar[0]));
    } else {
      int i;
      vpar[1] = 0;
      for (i = 2; i <= n; i++) vpar[i] = 1;
    }

    for (;;) {
      DEBUG(nautyextra_print_vpar(vpar,n));
      count_arrays++;
      if (! nautyextra_is_cyclic_vpar(vpar,n)) {
        assert(nautyextra_validate_vpar(vpar,n));
        count_trees++;

        int directed;
        for (directed = 0; directed <= 1; directed++) {
          int i;
          for (i = 0; i < num_reps[directed]; i++) {
            if (nautyextra_is_isomorphic_vpar(vpar, rep_vpar[directed][i], n,
                                              directed, NULL)) {
              goto ISOMORPHIC;
            }
          }
          assert(num_reps[directed] < max_reps);
          memcpy(rep_vpar[directed][num_reps[directed]], vpar,
                 (n+1)*sizeof(int));
          num_reps[directed]++;
        ISOMORPHIC:
          ;
        }
      }

      int pos = 2;
      for (;;) {
        if (pos > n) goto END_N;
        if (++vpar[pos] <= n) {
          DEBUG(printf("  inc low %d\n", pos));
          break;
        }
        vpar[pos] = (forest ? 0 : 1);
        pos++;
      }
    }

  END_N:
    printf("  count arrays %d trees %d\n", count_arrays, count_trees);
    printf("  isomorphic classes undirected %d directed %d\n",
           num_reps[0], num_reps[1]);
  }
}

void
try_isomorphism_pairwise (void)
{
  /*
    number of labelled trees up to isomorphism
    1 2 4 9 20 48
    A000081

    number of root=1 labelled trees in each of these isomorphism classes
    1
    1 2
    1 3 6 6
    1 4 12 12 12 12 24 24 24
    1 5 20 20 20 30 60 60 60 60 60 60 60 60 120 120 120 120 120 120
    1 6 30 30 30 60 60 90 120 120 120 120 120 120 120 180 180 180 360 360 360 360 360 360 360 360 360 360 360 360 360 360 360 360 360 360 720 720 720 720 720 720 720 720 720 720 720 720

    not in OEIS: 1,5,20,20,20,30,60,60,60,60,60,60,60,60,120,120,120,120,120,120
  */
  const int forest = 1;
  const int directed = 1;

  const int max_n = 7;   /* 8 too big for 32 bits */

  int n;
  for (n=0; n<=max_n; n++) {
    printf("n=%d\n", n);

    int limit_vpars = nautyextra_can_count(n,forest);
    DYNALLSTAT(int,vpars,vpars_sz);
    DYNALLOC1(int,vpars,vpars_sz, limit_vpars*(n+1), "malloc");
    printf("  limit_vpars %d\n", limit_vpars);

    int vec[max_n];
    nautyextra_can_init(vec,n, forest);
    int num_vpars = 0;
    do {
      DEBUG({ printf("vec "); int i;
              for (i = 0; i < n; i++) printf(" %d", vec[i]);
              printf("\n");
            });
      if (num_vpars >= limit_vpars) {
        printf("limit %d vpars reached\n", limit_vpars);
        abort();
      }
      int *vpar = vpars + num_vpars*(n+1);
      num_vpars++;
      nautyextra_from_upascend(vec,n,0, vpar);

      DEBUG(nautyextra_print_vpar(vpar,n));
      if (forest == -1) {
        assert (nautyextra_num_roots_vpar(vpar,n) <= 1);
        assert (n==0 || vpar[1] == 0);
      }
    } while (nautyextra_can_next(vec,n, forest));
    printf("  count vpars %d\n", num_vpars);

    int m = SETWORDSNEEDED(num_vpars);
    unsigned long long  iso_words
      = (unsigned long long) m * (unsigned long long) num_vpars;
    unsigned long long  iso_bytes = iso_words * sizeof(setword);
    printf("  m = %d size %Lu words %Lu bytes\n",
           num_vpars, iso_words, iso_bytes);
    if (iso_bytes > SIZE_MAX) {
      printf ("  too big\n");
      abort();
    }

    DYNALLSTAT(graph,iso, iso_sz);
    DYNALLOC2(graph,iso, iso_sz, m,num_vpars,"malloc");
    EMPTYGRAPH(iso,m,num_vpars);
    int i,j;
    for (i = 0; i < num_vpars; i++) {
      int *vpar1 = vpars + i*(n+1);
      for (j = 0; j < num_vpars; j++) {
        int *vpar2 = vpars + j*(n+1);
        int is_isomorphic = nautyextra_is_isomorphic_vpar(vpar1,vpar2,n,
                                                          directed, NULL);
        if (num_vpars <= 10000) {
          int is_isomorphic2 = nautyextra_is_isomorphic_vpar2(vpar1,n,
                                                              vpar2,n,
                                                              directed);
          assert (is_isomorphic == is_isomorphic2);
        }
        if (is_isomorphic)
          ADDONEARC(iso, i,j, m);
      }
    }

    printf("  reflexive\n");
    for (i = 0; i < num_vpars; i++)
      assert (ISELEMENT(GRAPHROW(iso,m,i),i));  /* reflexive */

    printf("  symmetric\n");
    for (i = 0; i < num_vpars; i++)
      for (j = 0; j < num_vpars; j++)
        assert (! ISELEMENT(GRAPHROW(iso,m,i),j)
                == ! ISELEMENT(GRAPHROW(iso,m,j),i));  /* symmetric */

    printf("  transitive\n");
    for (i = 0; i < num_vpars; i++) {
      for (j = 0; j < num_vpars; j++) {
        if (! ISELEMENT(GRAPHROW(iso,m,i),j)) continue;
        int k;
        for (k = 0; k < num_vpars; k++) {
          if (! ISELEMENT(GRAPHROW(iso,m,j),k)) continue;
          assert (ISELEMENT(GRAPHROW(iso,m,i),k));  /* transitive */
        }
      }
    }

    if (num_vpars <= 38)
      nautyextra_print_matrix((graph*)iso,m,num_vpars);

    printf("  classes\n");
    DYNALLSTAT(int,rep,rep_sz);
    DYNALLSTAT(int,repcount,repcount_sz);
    int num_reps = 0;
    for (i = 0; i < num_vpars; i++) {
      for (j = 0; j < num_reps; j++) {
        int r = rep[j];
        assert (0 <= r && r < num_vpars);
        if (ISELEMENT(GRAPHROW(iso,m,i),r)) {
          repcount[j]++;
          goto NEXT_REP;
        }
      }
      DYNREALLOC(int,rep,rep_sz, num_reps+1, "malloc");
      DYNREALLOC(int,repcount,repcount_sz, num_reps+1, "malloc");
      rep[num_reps] = i;
      repcount[num_reps] = 1;
      num_reps++;
    NEXT_REP:
      ;
    }

    qsort(repcount,num_reps,sizeof(repcount[0]),compare_int);

    printf("  number of classes %d\n", num_reps);
    printf("  counts ");
    for (i = 0; i < num_reps; i++)
      printf(" %d", repcount[i]);
    printf("\n");
  }
}

/* fprintf vpar[1..n] to file point fp in a form suitable for Pari/GP to
   read as a vector.  This means square brackets like [0,1,2,3].
   There are no newlines anywhere, so the vector is suitable for inclusion
   in a bigger line or expression.
*/
void
nautyextra_fprint_vpar_gp (FILE *fp, const int *vpar, int n)
{
  int i;
  fputc('[', fp);
  for (i = 1; i <= n; i++) {
    if (i != 1) fputc(',', fp);
    fprintf(fp,"%d",vpar[i]);
  }
  fputc(']', fp);
}

/* Run gp in a popen() output pipe and print to it some tests of
   vpar_is_isomorphic() to see that it matches nauty.
   Both directed and undirected (free tree) isomorphisms are tried.
*/
void
try_isomorphism_to_gp (void)
{
  int verbose = 1;
  const int min_n = 0;
  const int max_n = 12;
  const int forest = 1;

  FILE *fp = popen("gp -f --quiet","w");
  if (!fp) abort();
  setbuf(fp,NULL);

  if (verbose)
    fprintf(fp,"\\e\n");

  fprintf(fp,"default(recover,0)\n");
  fprintf(fp,"read(\"../../vpar/vpar.gp\");\n");
  fprintf(fp,"bad()=print(\"got  \"got);print(\"want \"want);vpar_view(vpar1);vpar_view(vpar2);print(\"bad free=\",free);error();\n");
  fprintf(fp,"count=0;\n");

  assert(max_n >= min_n);
  int n;
  for (n = min_n; n <= max_n; n++) {
    printf("sending n=%d\n", n);
    fprintf(fp,"print(\"n=%d\")\n", n);

    int vec1[max_n];
    nautyextra_can_init(vec1,n, forest);
    do {
      int vpar1[max_n+1];
      nautyextra_from_upascend(vec1,n,0, vpar1);

      int vec2[max_n];
      nautyextra_can_init(vec2,n, forest);
      do {
        int vpar2[max_n+1];
        nautyextra_from_upascend(vec2,n,0, vpar2);

        fprintf(fp,"vpar1=");
        nautyextra_fprint_vpar_gp(fp,vpar1,n);
        fprintf(fp,"%s\n", verbose?"":";");
        fprintf(fp,"vpar2=");
        nautyextra_fprint_vpar_gp(fp,vpar2,n);
        fprintf(fp,"%s\n", verbose?"":";");

        int free;
        for (free=0; free<=1; free++) {
          int directed = 1 - free;
          int want = nautyextra_is_isomorphic_vpar2 (vpar1,n, vpar2,n,
                                                     directed);
          fprintf(fp,"want=%d%s\n", want, verbose?"":";");
          fprintf(fp,"got=%s(vpar1,vpar2)%s\n",
                  free ? "vpar_is_isomorphic" : "vpar_is_free_isomorphic",
                  verbose?"":";");
          fprintf(fp,"if(got!=want,bad());\n");
          fprintf(fp,"count++;\n");
        }
      } while (nautyextra_can_next(vec2,n, forest));
    } while (nautyextra_can_next(vec1,n, forest));
  }
  fprintf(fp,"print(\"total count \"count);\n");
  pclose(fp);
}

static int lexmin_vpar[20];
int relabel_lexmin_temp_vpar[20][20];
int relabel_lexmin_recurse_n;
void
relabel_lexmin_recurse (int pos)
{
  int n = relabel_lexmin_recurse_n;
  if (pos >= n) {
    DEBUG(printf("  compare ");
          nautyextra_fprint_vpar_gp(stdout,relabel_lexmin_temp_vpar[pos-1],n);
          printf("\n"));
    const int *this_vpar = relabel_lexmin_temp_vpar[pos-1];
    if (memcmp(this_vpar+1, lexmin_vpar+1, n*sizeof(int)) < 0) {
      memcpy(lexmin_vpar+1, this_vpar+1, n*sizeof(int));
      DEBUG(printf(" new low "));
    }

  } else {
    int swap;
    for (swap=pos; swap<=n; swap++) {
      DEBUG(printf(" swap %d %d\n", pos, swap));
      nautyextra_relabel_swap_vpar_copy(relabel_lexmin_temp_vpar[pos-1], n,
                                        pos, swap,
                                        relabel_lexmin_temp_vpar[pos]);
      relabel_lexmin_recurse (pos+1);
    }
  }
}
void
relabel_lexmin (const int *vpar, int n)
{
  DEBUG(printf("relabel_lexmin() ");
        nautyextra_fprint_vpar_gp(stdout,vpar,n); printf("\n"));

  assert (n<20);
  memcpy (lexmin_vpar+1, vpar+1, n*sizeof(int));
  memcpy (relabel_lexmin_temp_vpar[0], vpar, (n+1)*sizeof(int));
  relabel_lexmin_recurse_n = n;
  relabel_lexmin_recurse(1);

  DEBUG(printf(" lexmin ");
        nautyextra_fprint_vpar_gp(stdout,lexmin_vpar,n); printf("\n"));
}

/* Run gp in a popen() output pipe and print to it some tests of
   vpar_to_lexmin() vs search here.
*/
void
try_relabel_lexmin_to_gp (void)
{
  int verbose = 0;
  const int min_n = 2;
  const int max_n = 12;
  const int forest = 1;

  FILE *fp = popen("gp -f --quiet","w");
  if (!fp) abort();
  setbuf(fp,NULL);

  if (verbose)
    fprintf(fp,"\\e\n");

  fprintf(fp,"default(recover,0)\n");
  fprintf(fp,"read(\"../../vpar/vpar.gp\"); read(\"../../vpar/vpar-wip.gp\");\n");
  fprintf(fp,"bad()=print(\"got  \"got);print(\"want \"want);vpar_view(vpar1);vpar_view(vpar2);print(\"bad free=\",free);error();\n");
  fprintf(fp,"count=0;\n");

  assert(max_n >= min_n);
  int n;
  for (n = min_n; n <= max_n; n++) {
    printf("sending n=%d\n", n);
    fprintf(fp,"print(\"n=%d\")\n", n);

    int vec[max_n];
    nautyextra_can_init(vec,n, forest);
    do {
      int vpar[max_n+1];
      nautyextra_from_upascend(vec,n,0, vpar);

      relabel_lexmin(vpar,n);
      fprintf(fp,"vpar=");
      nautyextra_fprint_vpar_gp(fp,vpar,n);
      fprintf(fp,"%s\n", verbose?"":";");

      fprintf(fp,"want=");
      nautyextra_fprint_vpar_gp(fp,lexmin_vpar,n);
      fprintf(fp,"%s\n", verbose?"":";");

      fprintf(fp,"got=vpar_to_lexmin(vpar)%s\n", verbose?"":";");
      fprintf(fp,"if(got!=want,bad());\n");
      fprintf(fp,"count++;\n");

    } while (nautyextra_can_next(vec,n, forest));
  }
  fprintf(fp,"print(\"total count \"count);\n");
  pclose(fp);
}


/* ------------------------------------------------------------------------- */
/*
  trees root=1 or root=any num_children sorted uniq
  1, 1, 2, 3, 5, 7, 11, 15, 22,
  A000041 partitions of n
  forests same but one sooner n
  1, 2, 3, 5, 7, 11, 15, 22,

  ----
  trees root=any num_children uniq
  1, 2, 6, 20, 70, 252, 924, 3432,    for n>=1
  A000984 central binomial(2*n,n)
  vector(10,n,binomial(2*(n-1),n-1))

  trees root=1 num_children uniq, same as forest but one earlier
  1, 1, 3, 10, 35, 126, 462, 1716, 6435,

  forests num_children uniq
  1, 3, 10, 35, 126, 462, 1716, 6435,
  A001700 binomial(2n+1, n+1)
  = A088218 total num leaves in rooted ordered trees n
  num ordered partitions of n into n parts, allowing zeros = binomial(2*n-1,n)

*/
void
try_num_children_duplication (void)
{
  int verbose = 0;
  const int max_n = 12;
  const int forest = -1;

  int n;
  for (n = 1; n <= max_n; n++) {
    const int seen_numberof = 10000;
    int seen_width = n+1;
    int seen_width_bytes = seen_width*sizeof(int);
    int seen[seen_width*seen_numberof];
    int seen_upto = 0;

    int vec[max_n];
    nautyextra_can_init(vec,n, forest);
    int count_all = 0;
    int count = 0;
    do {
      int vpar[max_n+1];
      nautyextra_from_upascend(vec,n,0, vpar);
      count_all++;

      int num_children[max_n+1];
      nautyextra_num_children_array_vpar(vpar,n, num_children);
      /* qsort(num_children,seen_width,sizeof(int),compare_int); */

      int i;
      for (i=0; i<seen_upto; i++)
        if (memcmp(seen+seen_width*i, num_children, seen_width_bytes) == 0)
          goto SEEN;
      count++;
      assert(seen_upto < seen_numberof);
      memcpy(seen+seen_width*seen_upto, num_children, seen_width_bytes);
      seen_upto++;
  SEEN:
      ;
  } while (nautyextra_can_next(vec,n, forest));
  if (verbose) {
    printf("count all %d\n", count_all);
  }
  printf("%d,\n", count);
}
}

/* ------------------------------------------------------------------------- */
/*
  recurrence in counts of subtrees, see seq-subtree-heights.gp

  trees:
  count 1, 1, 2, 5, 16, 61, 268, 1329, 7342, 44703, 297314,
  n =   1  2  3  4   5   6    7     8     9     10      11
  not in OEIS: 2, 5, 16, 61, 268, 1329, 7342, 44703, 297314,

  forests same since could take extra vertex above components

  what was this:
  count 1, 2, 6, 20, 80, 366, 1876, 10632,
*/
void
try_subtree_height_idseq (void)
{
  int verbose = 0;
  const int max_n = 12;
  const int forest = 1;

  int n;
  for (n = 1; n <= max_n; n++) {
    int vec[max_n];
    nautyextra_can_init(vec,n, forest);
    int count_all = 0;
    int count = 0;
    do {
      int vpar[max_n+1];
      nautyextra_from_upascend(vec,n,0, vpar);

      /* subtree_height always root last, so idseq must have root=n */
      if (forest == -1)
        nautyextra_reroot_vpar(vpar,n, n); /* root=n */

      if (verbose) {
        printf("vpar ");
        nautyextra_print_vpar(vpar,n);
      }
      count_all++;

      int seq[max_n+1];
      nautyextra_seq_subtree_height_vpar(vpar,n, seq);
      if (verbose) {
        printf(" seq ");
        nautyextra_print_array(seq,n+1);
      }

      int is_idperm = nautyextra_is_idperm(seq,n+1);
      if (verbose) {
        printf("  is_idperm %d\n", is_idperm);
      }

      count += is_idperm;
    } while (nautyextra_can_next(vec,n, forest));
    if (verbose) {
      printf("count all %d\n", count_all);
    }
    printf("%d,\n", count);
  }
}

/* ------------------------------------------------------------------------- */
/*
   not in OEIS: 1, 1, 2, 4, 9, 19, 44, 96, 220, 489,
   differences:
   not in OEIS:      1, 2, 5, 10, 25, 52, 124, 269

   not A247623 = 1,1,2,4,9,19,44,96,225,

   superseeker of 1, 2, 4, 9, 19, 44, 96, 220, 489 gave:
   SUGGESTION: LISTTOALGEQ FOUND ONE OR MORE ALGEBRAIC
   EQUATIONS SATISFIED BY THE GEN.  FN.
   WARNING: THESE MAY BE ONLY APPROXIMATIONS!
   Equation(s) and type(s) are:
   .                            2         3       4
   .     [16 - 32 a(n) + 24 a(n)  - 8 a(n)  + a(n) , lgdegf]

*/
void
try_subtree_sizes_duplication (void)
{
  int verbose = 0;
  const int max_n = 12;
  const int forest = -1;

  int n;
  for (n = 1; n <= max_n; n++) {
    const int seen_numberof = 10000;
    int seen_width = n+1;
    int seen_width_bytes = seen_width*sizeof(int);
    int seen[seen_width*seen_numberof];
    int seen_upto = 0;

    int vec[max_n];
    nautyextra_can_init(vec,n, forest);
    int count_all = 0;
    int count = 0;
    do {
      int vpar[max_n+1];
      nautyextra_from_upascend(vec,n,0, vpar);
      count_all++;

      int subtree_sizes[max_n+1];
      nautyextra_subtree_sizes_array_vpar(vpar,n, subtree_sizes);
      qsort(subtree_sizes,seen_width,sizeof(int),compare_int);

      int i;
      for (i=0; i<seen_upto; i++)
        if (memcmp(seen+seen_width*i, subtree_sizes, seen_width_bytes) == 0)
          goto SEEN;
      count++;
      assert(seen_upto < seen_numberof);
      memcpy(seen+seen_width*seen_upto, subtree_sizes, seen_width_bytes);
      seen_upto++;
    SEEN:
      ;
    } while (nautyextra_can_next(vec,n, forest));
    if (verbose) {
      printf("count all %d\n", count_all);
    }
    printf("%d,\n", count);
  }
}

/* ------------------------------------------------------------------------- */

void
try_bandwidth (void)
{
  int n = 3;
  NAUTYEXTRA_TYPE_ULL limit = nautyextra_factorial(n);
  NAUTYEXTRA_TYPE_ULL num;
  for (num = 0; num < limit; num++) {
    int perm[n+1];
    nautyextra_numtoperm (num,n,perm);
    nautyextra_print_array(perm,n);
  }
  printf("one\n");
  for (num = 0; num < limit; num++) {
    int perm[n+1];
    nautyextra_numtoperm1 (num,n+1,perm);
    nautyextra_print_array(perm,n+1);
  }
}

void
try_Ramsey_4_4 (void)
{
  const int n = 17;
  const int m = SETWORDSNEEDED(n);
  graph g[m*n];

  EMPTYGRAPH(g,m,n);
  const int deltas[] = {1,2,4,8,9,13,15,16};
  int i,j;
  for (i=0; i<n; i++) {
    for (j=0; j<numberof(deltas); j++) {
      int delta = deltas[j];
      int to = (i+delta) % n;
      /* printf ("%d to %d\n", i, to); */
      ADDONEEDGE(g, i,to, m);
    }
  }
  assert (loopcount(g,m,n) == 0);

  nautyextra_view(g,m,n);

  int clique_number = nautyextra_clique_number(g,m,n);
  printf("clique_number %d\n", clique_number);

  struct nautyextra_indnum_and_count info;
  nautyextra_indnum_and_count(g,m,n,&info);
  printf("indnum %d\n", info.indnum);
}

void
try_Mobius_ladder_8_delete_2 (void)
{
  const int n = 8;
  const int m = SETWORDSNEEDED(n);
  graph g[m*n];

  EMPTYGRAPH(g,m,n);
  int i;
  for (i = 0; i<=3; i++) {
    ADDONEEDGE(g, i,i+1, m);     /* sides */
    ADDONEEDGE(g, i+4,i+4+1, m);
  }
  ADDONEEDGE(g, 0,7, m);  /* ends */
  ADDONEEDGE(g, 4,3, m);

  /* rungs */
  /* ADDONEEDGE(g, 1,5, m); */
  /* ADDONEEDGE(g, 3,7, m); */
  ADDONEEDGE(g, 0,4, m);
  ADDONEEDGE(g, 2,6, m);

  nautyextra_view(g,m,n);

  int clique_number = nautyextra_clique_number(g,m,n);
  printf("clique_number %d\n", clique_number);

  struct nautyextra_indnum_and_count info;
  nautyextra_indnum_and_count(g,m,n,&info);
  printf("indnum %d count %Lu\n", info.indnum, info.count);
}

/* Count the number of degrees vectors occurring for labelled forests of n
   vertices.

   unsorted
   not in OEIS: 1, 1, 2, 7, 30, 136, 629, 2941, 13858,
   not in OEIS:       2, 7, 30, 136, 629, 2941, 13858,

   sorted
   n          = 0  1  2  3  4   5   6   7   8   9
   not in OEIS: 1, 1, 2, 3, 6, 10, 18, 29, 48, 74,
   not in OEIS:       2, 3, 6, 10, 18, 29, 48, 74,
   */
void
try_forests_degrees (void)
{
  const int forest = 1;
  const int max_n = 16;

  int n;
  for (n=0; n<=max_n; n++) {
    DEBUG(printf("n=%d\n", n));

    int seen_array_size = 100000;
    DYNALLSTAT(int,seen_array,seen_array_sz);
    DYNALLOC1(int,seen_array,seen_array_sz, seen_array_size*n,
              "try_forests_degrees");
    int count = 0;

    int vec[max_n];
    nautyextra_can_init(vec,n, forest);
    do {
      int vpar[max_n+1];
      nautyextra_from_upascend(vec,n,0, vpar);

      int degrees[max_n+1];
      nautyextra_degrees_array_vpar(vpar,n,degrees);
      DEBUG(printf("degrees "); nautyextra_print_array (degrees+1, n));

      if (0) {
        qsort(degrees+1,n,sizeof(int),compare_int);
        DEBUG(printf(" sorted "); nautyextra_print_array (degrees+1, n));
      }

      int i;
      for (i=0; i<count; i++) {
        DEBUG(printf(" cmp %d %p  ", i, seen_array + n*i);
              nautyextra_print_array (seen_array + n*i, n));
        if (memcmp(seen_array + n*i, degrees+1, sizeof(int)*n) == 0)
          goto SEEN;
      }
      assert(count < seen_array_size);
      memcpy (seen_array + n*count, degrees+1, sizeof(int)*n);
      DEBUG(printf(" new to %d %p  ", count, seen_array + n*count);
            nautyextra_print_array (seen_array + n*count, n));
      count++;
    SEEN:
      ;
    } while (nautyextra_can_next(vec,n, forest));

    printf("%d,\n", count);

    if (0 && n <= 3) {
      int i;
      for (i=0; i<count; i++) {
        nautyextra_print_array (seen_array + n*i, n);
      }
    }
  }
}


/* ------------------------------------------------------------------------- */

void
try_automorphism_free (void)
{
  static DEFAULTOPTIONS_SPARSEGRAPH(options);
  statsblk stats;

  SG_DECL(sg);
  SG_DECL(canon_sg);

  options.getcanon = TRUE;
  options.userautomproc = groupautomproc;
  options.userlevelproc = grouplevelproc;

  printf("make graph\n");

  nautyextra_add_edge_sg(&sg, 0,1);

  nautyextra_add_edge_sg(&sg, 0,2);
  nautyextra_add_edge_sg(&sg, 2,3);

  nautyextra_add_edge_sg(&sg, 0,4);
  nautyextra_add_edge_sg(&sg, 4,5);
  nautyextra_add_edge_sg(&sg, 5,6);

  printf("\n");
  printf("as generated\n");
  nautyextra_print_sg(&sg);
  /* nautyextra_view_sg (&sg); */

  /* DYNALLSTAT(int, degrees, degrees_sz); */
  /* DYNALLOC1(int, degrees, degrees_sz, sg.nv, "malloc"); */
  /* nautyextra_degrees_array_sg(&sg); */
  /* printf("degrees "); nautyextra_print_array (degrees, sg.nv); */

  int num_vertices = sg.nv;
  DYNALLSTAT(int, lab,    lab_sz);
  DYNALLSTAT(int, ptn,    ptn_sz);
  DYNALLSTAT(int, orbits, orbits_sz);
  DYNALLSTAT(int, map,    map_sz);
  DYNALLOC1(int, lab,    lab_sz,    num_vertices, "malloc");
  DYNALLOC1(int, ptn,    ptn_sz,    num_vertices, "malloc");
  DYNALLOC1(int, orbits, orbits_sz, num_vertices, "malloc");
  DYNALLOC1(int, map,    map_sz,    num_vertices, "malloc");

  printf("\nlabel sg\n");
  sparsenauty(&sg,lab,ptn,orbits,&options,&stats,&canon_sg);

  printf("canonical\n");
  printf("lab "); nautyextra_print_array (lab, sg.nv);
  /* nautyextra_degrees_array_sg(&canon_sg); */
  /* printf("degrees "); nautyextra_print_array (degrees, sg.nv); */
  /* nautyextra_print_sg(&canon_sg); */
  /* nautyextra_view_sg (&canon_sg); */

  sortlists_sg(&sg);

  printf("putgraph_sg()\n");
  putgraph_sg(stdout, &sg, 79);  /* line length */

  printf("\n");
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&sg);
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&canon_sg);

  printf("\n");
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&sg);
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&canon_sg);

  printf("aresame  %d\n", aresame_sg(&sg,&canon_sg));

  printf ("\nGenerators\n");
  printf ("automorphism group size = ");
  writegroupsize(stdout, stats.grpsize1, stats.grpsize2);
  printf ("\n");
  printf ("num orbits    %d\n", stats.numorbits);
  printf ("  orbits      ");
  int i;
  for (i = 0; i < sg.nv; i++) printf(" %d",orbits[i]);
  printf("\n");
  printf ("  v =         ");
  for (i = 0; i < sg.nv; i++) printf(" %d",i);
  printf("\n");
  printf ("errstatus     %d\n", stats.errstatus);
  printf ("numnodes      %lu\n", stats.numnodes);
  printf ("numbadleaves  %lu\n", stats.numbadleaves);
  printf ("maxlevel      %d\n", stats.maxlevel);

  {
    grouprec *group = groupptr(FALSE);
    if (! group) {
      printf("no grouprec\n");
    } else {
      printf("grouprec %p\n", group);
      makecosetreps(group);  /* ready for allgroup() */
      allgroup(group,writeautom);
    }
  }

  const int n = 7;
  assert (sg.nv == n);
  int m = SETWORDSNEEDED(n);
  graph g[m*n];
  int m_out;
  sg_to_nauty(&sg,g,m,&m_out);

  struct nautyextra_similar_info info;
  nautyextra_similar(g,m,n, &info);

  printf("similar:\n");
  nautyextra_print_matrix(info.similar,m,n);
  printf("numorbits %d\n", info.numorbits);
  printf("orbit_max %d\n", info.orbit_max);
  printf("\n");

  printf("removal similar:\n");
  nautyextra_print_matrix(info.removal_similar,m,n);
  printf("\n");

  printf("pseudosimilar:\n");
  nautyextra_print_matrix(info.pseudosimilar,m,n);
  printf("pseudosimilar_clique_size %d\n", info.pseudosimilar_clique_size);

  {
    int i,j;
    for (i = 0; i < sg.nv; i++)
      for (j = i+1; j < sg.nv; j++)
        if (ISELEMENT(GRAPHROW(info.pseudosimilar,m,i),j))
          printf("pseudosimilar %d -- %d\n", i,j);
  }
  printf("vertex  ");
  for (i = 0; i < sg.nv; i++) printf(" %d",i);
  printf("\n");
  printf("degree  ");
  for (i = 0; i < sg.nv; i++) printf(" %d",sg.d[i]);
  printf("\n");
  printf("adj2s   ");
  for (i = 0; i < sg.nv; i++) {
    int count_neighbour_2s = 0;
    int pos;
    for (pos = 0; pos < sg.d[i]; pos++) {
      int j = sg.e[sg.v[i]+pos];
      count_neighbour_2s += (sg.d[j]==2);
    }
    printf(" %d",count_neighbour_2s);
  }
  printf("\n");

  printf("adj3s   ");
  for (i = 0; i < sg.nv; i++) {
    int count_neighbour_3s = 0;
    int pos;
    for (pos = 0; pos < sg.d[i]; pos++) {
      int j = sg.e[sg.v[i]+pos];
      count_neighbour_3s += (sg.d[j]==3);
    }
    printf(" %d",count_neighbour_3s);
  }
  printf("\n");
}







/* ------------------------------------------------------------------------- */

void
try_empty (void)
{
  SG_DECL(sg);
  SG_INIT(sg);
  printf("%s", sgtos6(&sg));
  nautyextra_ensure_vertex_sg (&sg,0);
  printf("%s", sgtos6(&sg));
  nautyextra_add_edge_sg(&sg, 0,0);
  printf("%s", sgtos6(&sg));
  nautyextra_add_edge_sg(&sg, 0,0);
  printf("%s", sgtos6(&sg));
  nautyextra_add_edge_sg(&sg, 0,0);
  printf("%s", sgtos6(&sg));
  nautyextra_add_edge_sg(&sg, 0,0);
  printf("%s", sgtos6(&sg));

  int n; size_t e;
  stringcounts(":@?N",&n,&e);
  printf("n=%d e=%u\n", n,e);
}

void
show_config (void)
{
#if HAVE_CLZ
  printf ("HAVE_CLZ  yes\n");
#else
  printf ("HAVE_CLZ  no\n");
#endif
}

void
try_digraph6_header(void)
{
  printf("NAUTYVERSIONID %d\n", NAUTYVERSIONID);
  {
    const char *str = ">>digraph6<<&IOB???????????????\n";
    str = "&?\n";
    str = EDGECODE_HEADER;
    str = ">>digraph6<<&?\n";
    FILE *fp = fopen("/tmp/foo.d6","w");
    if (!fp) abort();
    if (fputs(str, fp) == EOF) abort();
    if (fclose(fp) != 0) abort();
  }
  int codetype = 9999;
  FILE *fp = opengraphfile("/tmp/foo.d6", &codetype, 0, 0);
  printf("fp=%p  codetype=%d\n", fp, codetype);
}
void
try_CoJ(void)
{
  printf("NAUTYVERSIONID %d\n", NAUTYVERSIONID);

  /*
    copyg is cat when input format == output format

    echo ':CoJ' | nauty-copyg -g | nauty-copyg -s
    echo ':COJ' | nauty-copyg -g | nauty-copyg -s
  */

  const char *str = ":CoJ\n";
  str = ":CoJ\n";
  str = ":COJ\n";
  str = ":GoBN_\n";
  printf("%s", str);

  FILE *fp = fopen("/tmp/foo.s6","w+");
  if (!fp) abort();
  if (fputs(str, fp) == EOF) abort();
  if (fseek(fp,0,SEEK_SET) != 0) abort();

  int n_max = 32;
  int m_max = SETWORDSNEEDED(n_max);
  graph g[64];
  EMPTYGRAPH(g,m_max,n_max);
  int n,m, digraph=99;
  graph *ret;
  if (1) {
    graph *gprev = NULL;
    int nprev=1, mprev=1;
    ret = readgg_inc(fp,g,1,&m,&n, gprev,mprev,nprev, &digraph);
  } else {
    ret = readg(fp,g,1,&m,&n);
  }
  if (ret == NULL) abort();
  printf("ret=%p n=%d m=%d digraph=%d\n", ret, n,m, digraph);
  nautyextra_print_matrix(g,m,n);

  writes6(stdout,g,m,n);
}
void
try_digraph_direction(void)
{
  printf("NAUTYVERSIONID %d\n", NAUTYVERSIONID);

  int n = 5;
  int m = SETWORDSNEEDED(n);
  graph g[n];
  EMPTYGRAPH(g,m,n);
  ADDONEARC(g, 0,2, m);
  writed6(stdout, g,m,n);

  EMPTYGRAPH(g,m,n);
  ADDONEARC(g, 2,0, m);
  writed6(stdout, g,m,n);
}

/* ------------------------------------------------------------------------- */

void
try_isautom_sg(void)
{
  printf("NAUTYVERSIONID %d\n", NAUTYVERSIONID);

  int n = 5;
  SG_DECL(sg);
  SG_DECL(sgR);
  SG_INIT(sg);
  SG_INIT(sgR);

  nautyextra_add_edge_sg(&sg,0,1);
  nautyextra_add_edge_sg(&sg,1,2);
  nautyextra_add_edge_sg(&sg,2,3);
  sortlists_sg(&sg);
  printf("sg "); nautyextra_print_sg(&sg);

  /* NAUTYEXTRA_TYPE_ULL limit = nautyextra_factorial(n); */
  NAUTYEXTRA_TYPE_ULL i;
  for (i=1; i<=8; i++) {
    int perm[n];
    nautyextra_numtoperm (i,n,perm);
    printf("i=%Lu perm ",i); nautyextra_print_array(perm,n);

    copy_sg(&sg,&sgR);
    relabel_sg(&sgR, NULL, perm, NULL);
    sortlists_sg(&sgR);

    printf(" sg  %s", sgtos6(&sg));
    printf(" sgR %s", sgtos6(&sgR));

    int same = aresame_sg(&sg,&sgR);
    int aut  = isautom_sg((graph*)&sg,perm, 0,0,0);
    printf("%d %d\n", same, aut);
  }
}

/* ------------------------------------------------------------------------- */

void
try_near_perfect_matching (void)
{
  int verbose = 0;
  const int max_n = 12;
  const int forest = 1;
  const int near = 0;

  int n;
  for (n = 1; n <= max_n; n++) {
    int vec[max_n];
    nautyextra_can_init(vec,n, forest);
    int count_all = 0;
    int count = 0;
    if ((n&1) == near) {
      int target_matchnum = n>>1;
      do {
        int vpar[max_n+1];
        nautyextra_from_upascend(vec,n,0, vpar);
        count_all++;

        if (near) {
          int matchnum = nautyextra_matchnum_vpar(vpar,n);
          count += (matchnum == target_matchnum);
        } else {
          count += nautyextra_has_perfect_matching_vpar(vpar,n);
        }

      } while (nautyextra_can_next(vec,n, forest));
    }
    if (verbose) {
      printf("count all %d\n", count_all);
    }
    printf(" %d,", count);
  }
  printf("\n");
}

/* ------------------------------------------------------------------------- */
/* depths==vpar
   1, 2, 5, 14, 43, 144, 523, 2048, 8597, 38486,
   A047970 antidiagonal sums of nexus numbers A047969
 */
void
try_depths_eq_vpar (void)
{
  int verbose = 0;
  const int max_n = 12;
  const int forest = 1;

  int n;
  for (n = 1; n <= max_n; n++) {
    int vec[max_n];
    nautyextra_can_init(vec,n, forest);
    int count = 0;
    do {
      int vpar[max_n+1];
      nautyextra_from_upascend(vec,n,0, vpar);

      int depths[max_n+1];
      nautyextra_depths_vpar(vpar,n,depths);
      if (memcmp(vpar+1,depths+1,n*sizeof(int))==0) {
        count++;
        if (verbose) {
        }
      }
    } while (nautyextra_can_next(vec,n, forest));

    printf(" %d,", count);
  }
  printf("\n");
}

/* ------------------------------------------------------------------------- */

void
try_writegroupsize(void)
{
  int n;
  for (n = 5; n <= 25; n += 5) {
    int m = SETWORDSNEEDED(n);
    graph g[n];
    EMPTYGRAPH(g,m,n);

    DEFAULTOPTIONS_GRAPH(options);
    statsblk stats;
    int lab[n];
    int ptn[n];
    int orbits[n];
    densenauty ((graph*)g, lab, ptn, orbits, &options, &stats, m,n, NULL);
    printf("n=%d writegroupsize() ", n);
    writegroupsize(stdout, stats.grpsize1, stats.grpsize2);
    printf ("\n");
  }
}

/* ------------------------------------------------------------------------- */

int
main (void)
{
  setbuf(stdout,NULL);
  try_writegroupsize();
  try_relabel_lexmin_to_gp();
  return 0;
  try_depths_eq_vpar();
  try_near_perfect_matching();
  try_isautom_sg();
  try_digraph_direction();
  try_CoJ();
  try_digraph6_header();
  try_forests_degrees();
  try_automorphism_free();
  try_isomorphism_to_gp();
  try_subtree_sizes_duplication();
  try_Mobius_ladder_8_delete_2();
  try_Ramsey_4_4();
  try_num_children_duplication();
  try_subtree_height_idseq();
  try_pseudosimilar();
  try_bandwidth();
  try_isomorphism_pairwise();
  try_isomorphism_classes();
  try_empty();
  try_generalized_Petersen_diameter();
  try_generalized_Petersen_sets();
  try_generalized_Petersen_count();
  try_cycle_totdomnum();
  try_Petersen_7_2();
  try_indomnum_n9();
  show_config();
  return 0;
}
