/* Copyright 2018 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Run some endless tests of nautyextra_disjoint_domnum_vpar() and
   nautyextra_disjoint_domnum_and_count_vpar(), using trees bigger than
   really want in the general tests/test-nautyextra.c.
*/

#include <assert.h>
#include <math.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int option_verbose = 0;

int count_total = 0;
int count_bad = 0;
const int max_bad = 1;

/*---------------------------------------------------------------------------*/

#define numberof(array)  (sizeof(array)/sizeof((array)[0]))

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))


/*---------------------------------------------------------------------------*/
/* Disjoint Dominating Sets Helpers */

void
test_disjoint_domnum_and_count_by_all_sets (const sparsegraph *sg,
                                            struct nautyextra_disjoint_domnum_and_count *info)
{
  int n = sg->nv;
  assert (n >= 0);
  if (n == 0) {
    info->disjoint_domnum = 0;
    info->count = 1;
    return;
  }

  int disjoint_domnum = n+1;
  unsigned long count = 0;

  int size = 0;
  setword ss1 = 0;
  setword ss2 = 0;
  setword lowbit = NAUTYEXTRA_SETWORD_HIGHBIT >> (n-1);
  for (;;) {
    setword bit = NAUTYEXTRA_SETWORD_HIGHBIT;
    for (;;) {
      if (ss1 & bit) {
        ss1 ^= bit;    /* 1 -> 2 */
        ss2 ^= bit;
        break;
      }
      if (ss2 & bit) {
        ss2 ^= bit;    /* 2 -> 0 and step down */
        size--;
        if (bit == lowbit) {
          if (count == 0) disjoint_domnum = -1;
          info->disjoint_domnum = disjoint_domnum;
          assert ((count&1) == 0);
          info->count = count >> 1;
          return;
        }
        bit >>= 1;
        continue;
      }
      ss1 ^= bit;      /* 0 -> 1 */
      size++;
      break;
    }

    assert ((ss1 & ss2) == 0);
    if (nautyextra_is_domset_sg(sg,&ss1)
        && nautyextra_is_domset_sg(sg,&ss2)) {
      if (size < disjoint_domnum) {
        disjoint_domnum = size;
        count = 1;
      } else if (size == disjoint_domnum) {
        count++;
      }
    }
  }
}

#define GETKRAN_RANGE(lo,hi, var)                                       \
  do {                                                                  \
    int __GETKRAN_RANGE__lo = (lo);                                     \
    int __GETKRAN_RANGE__range = (hi) - __GETKRAN_RANGE__lo + 1;        \
    int __GETKRAN_RANGE__random;                                        \
    GETKRAN(__GETKRAN_RANGE__range, __GETKRAN_RANGE__random);           \
    (var) = __GETKRAN_RANGE__random + __GETKRAN_RANGE__lo;              \
  } while (0)

int
main (void)
{
  const int lo = 6;
  const int hi = 16;

  setbuf(stdout,NULL);
  nautyextra_want_progress = 1;
  int vpar[hi+1];
  SG_DECL(sg);
  SG_INIT(sg);

  for (;; count_total++) {
    if (NAUTYEXTRA_PROGRESS)
      printf("count_total %d\r", count_total);

    int n, forest;
    GETKRAN_RANGE(lo,hi, n);
    GETKRAN(2, forest);
    nautyextra_make_random_vpar(vpar,n, forest);
    nautyextra_vpar_to_sg(vpar,n, &sg, 0);

    /* nautyextra_disjoint_domnum_vpar()
       nautyextra_disjoint_domnum_and_count_vpar()
    */
    struct nautyextra_disjoint_domnum_and_count info_by_test;
    test_disjoint_domnum_and_count_by_all_sets(&sg,&info_by_test);

    int by_vpar = nautyextra_disjoint_domnum_vpar(vpar,n);
    if (by_vpar != info_by_test.disjoint_domnum) {
      printf("n=%d disjoint domnum\n", n);
      printf("  by_vpar  %d\n", by_vpar);
      printf("  by_test  %d\n", info_by_test.disjoint_domnum);
      nautyextra_view_sg(&sg);
      if (++count_bad >= max_bad) assert(0);
    }

    struct nautyextra_disjoint_domnum_and_count info_by_vpar;
    nautyextra_disjoint_domnum_and_count_vpar(vpar,n,&info_by_vpar);
    if (info_by_vpar.disjoint_domnum != info_by_test.disjoint_domnum
        || info_by_vpar.count        != info_by_test.count) {
      printf("n=%d disjoint domnum\n", n);
      printf("  info_by_vpar  %d %Lu\n",
             info_by_vpar.disjoint_domnum, info_by_vpar.count);
      printf("  info_by_test  %d %Lu\n",
             info_by_test.disjoint_domnum, info_by_test.count);
      nautyextra_view_sg(&sg);
      if (++count_bad >= max_bad) assert(0);
    }
  }
}
