/* Copyright 2017 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Search for graphs or trees with a particular mean distance between
   vertices as a fraction of the diameter (that being the longest distance
   between vertices).

   So Wiener index W, seek W/(pairs*diameter) = frac.
 */

/* #define _GNU_SOURCE */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif


int option_terminal = 0;
int option_tree = 1;         /* 0=all graphs, 1=trees only, -1=must cycle */
int option_bipartite = 0;
int option_clawfree = 1;
int option_num = 1;
int option_den = 2;
int option_min_vertices = 2;
int option_max_vertices = 32;
int option_verbose = 0;
int option_view = 1;
int option_count_found_limit = 30;
int option_view_found_limit = 5;

int my_geng_main(int argc, char**argv);
int my_gentreeg_main(int argc, char**argv);

int count_total = 0;
int count_found = 0;

SG_DECL(sg);

void
my_gentreeg_outproc (FILE *outfile, int *vpar, int n)
{
  if (option_clawfree && ! nautyextra_is_clawfree_vpar(vpar,n)) {
    return;
  }

  if (count_found > option_count_found_limit) {
    printf ("... and more\n");
    exit(0);
  }

  count_total++;
  if (option_verbose) {
    printf("tree n=%d  %p\n", n, vpar);
    /* nautyextra_print_matrix(g,m,n); */
  }

  struct nautyextra_Wiener_info info;
  nautyextra_Wiener_index_vpar(vpar,n, option_terminal, &info);
  unsigned long long W = info.W;
  if (W == 0) return;

  int diameter = nautyextra_diameter_vpar(vpar,n);
  int num_pairs = n * (n-1) / 2;
  unsigned long long divisor = (unsigned long long) diameter * num_pairs;

  if (W*option_den == divisor*option_num) {
    count_found++;
    double mean_dist = (double) W / num_pairs;
    double mean_of_diam = (double) W / divisor;

    printf("n=%d found mean %.2f of diam %.6g  W=%Lu num_pairs=%d diameter=%d\n",
           n, mean_dist, mean_of_diam, W, num_pairs, diameter);

    nautyextra_vpar_to_sg(vpar,n, &sg, 0); /* undirected */
    SG_DECL(canon_sg);
    SG_INIT(canon_sg);
    nautyextra_canonise_sg(&sg,&canon_sg);
    printf("%s",GRAPH6_HEADER);
    writeg6_sg(stdout,&canon_sg);
    putgraph_sg(stdout,&canon_sg,-1);
    SG_FREE(canon_sg);

    if (n < 7 && count_found < option_view_found_limit) {
      if (option_view) nautyextra_view_sg(&sg);
    }
  }
}

void
my_geng_outproc (FILE *outfile, graph *g, int n)
{
  int m = SETWORDSNEEDED(n);
  int is_cyclic = -1;

  if (option_tree<0) {
    is_cyclic = nautyextra_is_cyclic(g,m,n);
    if (! is_cyclic) return;
  }
  if (option_clawfree && ! nautyextra_is_clawfree(g,m,n)) {
    return;
  }

  count_total++;
  if (option_verbose) {
    printf("graph n=%d  %p\n", n, g);
    nautyextra_print_matrix(g,m,n);
  }

  unsigned long long W;
  struct nautyextra_Wiener_info info;
  nautyextra_Wiener_index(g,m,n, option_terminal, &info);
  W = info.W;
  if (W == 0) return;

  int radius, diameter;
  diamstats(g, m,n, &radius, &diameter);

  int num_pairs = n * (n-1) / 2;
  unsigned long long divisor = (unsigned long long) diameter * num_pairs;

  /* W/divisor == num/den
     W*den == num*divisor */
  if (W*option_den == divisor*option_num) {
    if (count_found > option_count_found_limit) {
      printf ("... and more\n");
      exit(0);
    }

    if (is_cyclic < 0)
      is_cyclic = nautyextra_is_cyclic(g,m,n);

    count_found++;
    double mean_dist = (double) W / num_pairs;
    double mean_of_diam = (double) W / divisor;

    printf("n=%d found mean %.2f of diam %.6g  W=%Lu num_pairs=%d diameter=%d  cyclic=%d\n",
           n, mean_dist, mean_of_diam, W, num_pairs, diameter, is_cyclic);
    /* putgraph(stdout,g,-1,m,n); */
    printf("%s",GRAPH6_HEADER);
    writeg6(stdout,g,m,n);

    if (n < 7 || count_found < 6) {
      if (option_view) nautyextra_view(g,m,n);
    }
  }
}


int
main (void)
{
  setbuf(stdout,NULL);
  SG_INIT(sg);

  if (! getenv("DISPLAY"))
    option_view = 0;

  /* printf("NAUTYVERSIONID = %d\n", NAUTYVERSIONID); */

  if (option_max_vertices < option_min_vertices) {
    printf ("max_vertices < min_vertices\n");
    abort();
  }
#if MAXN
  if (option_max_vertices > MAXN) {
    printf ("max_vertices %d bigger than MAXN %d\n", option_max_vertices, MAXN);
    abort();
  }
#endif

  int  num_vertices;
  for (num_vertices = option_min_vertices;
       num_vertices <= option_max_vertices;
       num_vertices++) {
    printf("\nn=%d\n", num_vertices);

    {
      int n = num_vertices;
      int m = SETWORDSNEEDED(n);
      nauty_check(WORDSIZE,m,n,NAUTYVERSIONID);
    }

    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);
    int geng_argc = 0;
    char *geng_argv[10];
    geng_argv[geng_argc++] = (option_tree>0 ? "nauty-gentreeg" : "nauty-geng");
    geng_argv[geng_argc++] = "-q";     /* quiet */
    if (! option_tree)
      geng_argv[geng_argc++] = "-c";     /* connected graphs */

    if (option_bipartite) {
      if (option_tree>0) {
        printf ("all trees are bipartite anyway\n");
        abort();
      }
      geng_argv[geng_argc++] = "-b";     /* bipartite graphs */
    }

    geng_argv[geng_argc++] = num_vertices_str;
    geng_argv[geng_argc] = NULL;

    if (option_tree>0)
      my_gentreeg_main(geng_argc,geng_argv);
    else
      my_geng_main(geng_argc,geng_argv);
  }
  printf("count total %d found %d\n", count_total, count_found);

  return 0;
}
